//----------------------------------------------------------------------------------------
// mesh_container.cpp: Contains definition for the mesh container
//----------------------------------------------------------------------------------------

#include <cassert>
#include <iostream>
#include <fstream>
#include <sstream>
// #include <cstring>
// #include <cstdio>
#include <string>
#include <iomanip>
#include <utility>
#include <exception>
#include <vector>
#include <cmath>
#include "general.H"
#include "mesh_container.H"
#include "pde.H"
#include "error_computation.H"
#include "utility.H"
#if defined(TEST_KEY) || defined(SUPERCRITICAL_KEY)
#include "test.H"
#endif

extern PROBLEM_DESCRIPTION problem_info;  // Storage for physical parameters
extern QUAD_CONTAINER quadrature_index;   // Quadrature objects
extern BASIS_2D *basis_2d[max_order+1];   // 2D basis objects
extern BASIS_REF *basis_ref[max_order+1]; // 3D basis objects for parallel prisms
extern MESH mesh;                         // Mesh object
extern DOF var[d_var];                    // Global storage for degrees of freedom for unknowns
extern DOF var_rhs[d_var];                // Global storage for degrees of freedom for the right-hand side
extern DOF var_flux[d_var][d_space];      // Global storage for degrees of freedom for fluxes


//----------------------------------------------------------------------------------------
// MULTI_DEQUE template class: A list class containing objects of various subtypes
//                             derived from a common base class
//----------------------------------------------------------------------------------------

// Overloaded push_back function for EDGE_2D_SW type
// Parameters: edge - entry to be added to the list
template<>
void MULTI_DEQUE<EDGE_2D_SW>::push_back(const EDGE_2D_SW& edge)
{
  switch(edge.Type())
  {
  case EDGE_2D_SW::INTERIOR:
    push_back_index(mesh.edge_interior_list, (const EDGE_INTERIOR&)edge);
    break;
  case EDGE_2D_SW::LAND:
    push_back_index(mesh.edge_land_list, (const EDGE_LAND&)edge);
    break;
  case EDGE_2D_SW::RIVER:
    push_back_index(mesh.edge_river_list, (const EDGE_RIVER&)edge);
    break;
  case EDGE_2D_SW::SEA:
    push_back_index(mesh.edge_sea_list, (const EDGE_SEA&)edge);
    break;
  case EDGE_2D_SW::RADIATION:
    push_back_index(mesh.edge_radiation_list, (const EDGE_RADIATION&)edge);
    break;
#ifdef TEST_KEY
  case EDGE_2D_SW::UNKNOWN:
    push_back_index(mesh.edge_unknown_list, (const EDGE_UNKNOWN&)edge);
    break;
#endif
  default: std::cerr << "File mesh_container.cpp: Unknown edge type!\n";
    std::exit(1);
  }
}


// Overloaded push_back function for FACE_HORIZ_SW type
// Parameters: face - entry to be added to the list
template<>
void MULTI_DEQUE<FACE_HORIZ_SW>::push_back(const FACE_HORIZ_SW& face)
{
  switch(face.Type())
  {
  case FACE_HORIZ_SW::SURFACE:
    push_back_index(mesh.face_horiz_surface_list, (const FACE_HORIZ_SURFACE&)face);
    break;
  case FACE_HORIZ_SW::SURFACE_STRESS:
    push_back_index(mesh.face_horiz_surface_stress_list, (const FACE_HORIZ_SURFACE_STRESS&)face);
    break;
  case FACE_HORIZ_SW::INTERIOR:
    push_back_index(mesh.face_horiz_interior_list, (const FACE_HORIZ_INTERIOR&)face);
    break;
  case FACE_HORIZ_SW::BOTTOM_NO_SLIP:
    push_back_index(mesh.face_horiz_bottom_no_slip_list, (const FACE_HORIZ_BOTTOM_NO_SLIP&)face);
    break;
  case FACE_HORIZ_SW::BOTTOM_NO_DRAG:
    push_back_index(mesh.face_horiz_bottom_no_drag_list, (const FACE_HORIZ_BOTTOM_NO_DRAG&)face);
    break;
  case FACE_HORIZ_SW::BOTTOM_QUAD_DRAG:
    push_back_index(mesh.face_horiz_bottom_quad_drag_list, (const FACE_HORIZ_BOTTOM_QUAD_DRAG&)face);
    break;
  case FACE_HORIZ_SW::BOTTOM_LIN_DRAG:
    push_back_index(mesh.face_horiz_bottom_lin_drag_list, (const FACE_HORIZ_BOTTOM_LIN_DRAG&)face);
    break;
#ifdef TEST_KEY
  case FACE_HORIZ_SW::BOTTOM_TEST:
    push_back_index(mesh.face_horiz_bottom_test_list, (const FACE_HORIZ_BOTTOM_TEST&)face);
    break;
#endif
  default: std::cerr << "File mesh_container.cpp: Unknown horizonal face type!\n";
    std::exit(1);
  }
}


// Overloaded push_back function for FACE_VERT_SW type
// Parameters: face - entry to be added to the list
template<>
void MULTI_DEQUE<FACE_VERT_SW>::push_back(const FACE_VERT_SW& face)
{
  switch(face.Type())
  {
  case EDGE_2D_SW::INTERIOR: 
    push_back_index(mesh.face_vert_interior_list, (const FACE_VERT_INTERIOR&)face);
    break;
  case EDGE_2D_SW::LAND: 
    push_back_index(mesh.face_vert_land_list, (const FACE_VERT_LAND&)face);
    break;
  case EDGE_2D_SW::SEA: 
    push_back_index(mesh.face_vert_sea_list, (const FACE_VERT_SEA&)face);
    break;
  case EDGE_2D_SW::RIVER: 
    push_back_index(mesh.face_vert_river_list, (const FACE_VERT_RIVER&)face);
    break;
  case EDGE_2D_SW::RADIATION: 
    push_back_index(mesh.face_vert_radiation_list, (const FACE_VERT_RADIATION&)face);
    break;
#ifdef TEST_KEY
  case EDGE_2D_SW::UNKNOWN: 
    push_back_index(mesh.face_vert_unknown_list, (const FACE_VERT_UNKNOWN&)face);
    break;
#endif
  default: std::cerr << "File mesh_container.cpp: Unknown vertical face type!\n";
    std::exit(1);
  }
}


// Overloaded push_back function for PRISM_SW type
// Parameters: prism - entry to be added to the list
template<>
void MULTI_DEQUE<PRISM_SW>::push_back(const PRISM_SW& prism)
{
  switch(prism.Type())
  {
  case PRISM_SW::PARALLEL:
    push_back_index(mesh.prism_parallel_list, (const PRISM_PARALLEL&)prism);
    break;
  case PRISM_SW::NONPARALLEL:
    push_back_index(mesh.prism_nonparallel_list, (const PRISM_NONPARALLEL&)prism);
    break;
  default: std::cerr << "File mesh_container.cpp: Unknown prism type!\n";
    std::exit(1);
  }
}



//----------------------------------------------------------------------------------------
// MESH class: A list class for storage of mesh components
//----------------------------------------------------------------------------------------

// Input 2D mesh from a file (ADCIRC format)
// Parameters: grid_coord - coordinates file (fort.14); grid_connect - connectivity file (fort.17);
//             max_depth - maximum admissible bathymetric depth
void MESH::Input2DMesh(const char* grid_coord, const char* grid_connect, real& max_depth)
{
  std::ifstream g_coord(grid_coord), g_connect(grid_connect);

  if (!(g_coord&&g_connect))
  {
    std::cerr << "File mesh_container.cpp: Grid input files are not found!" << std::endl;
    std::exit(1);
  }

  try
  {
    std::size_t n_elements, n_edges, n_nodes;

    // Input from "grid_coord" file
    g_coord >> std::ws;
    g_coord.get(problem_info.grid_desc, sizeof(problem_info.grid_desc), '\n');
    g_coord.ignore(256, '\n');
    g_coord >> n_elements >> n_nodes;
    g_coord.ignore(256, '\n');


    // Input nodal coordinates
    for (std::size_t i=0; i<n_nodes; i++) 
    {
      int id;
      real coord_tmp[d_space_2d];
      NODE_2D_SW node;

      // Compute node coordinates (cartesian or spherical)
      switch (problem_info.coord_system)
      {
      case PROBLEM_DESCRIPTION::CARTESIAN: 
        g_coord >> id >> node.Coord(0) >> node.Coord(1) >> node.Depth();
        problem_info.CartesianToSpherical(node.Coord(), coord_tmp);
        node.SpherCoord(0)=coord_tmp[0];
        node.SpherCoord(1)=coord_tmp[1];
        break;
      case PROBLEM_DESCRIPTION::SPHERICAL: 
        g_coord >> id >> node.SpherCoord(0) >> node.SpherCoord(1) >> node.Depth();
        node.SpherCoord(0)*=problem_info.pi/180.;
        node.SpherCoord(1)*=problem_info.pi/180.;
        problem_info.SphericalToCartesian(node.SpherCoord(), coord_tmp);
        node.Coord(0)=coord_tmp[0];
        node.Coord(1)=coord_tmp[1];
        break;
      }
      if (id!=i+1)
      {
        std::cerr << "File mesh_container.cpp: Node ordering error at node: " << i << std::endl;
        std::exit(1);
      }
      g_coord.ignore(256, '\n');
      node.ID()=i;
      // Check minimal depth condition
      if (node.Depth() < problem_info.min_bathym_depth)
        node.Depth()=problem_info.min_bathym_depth;
      max_depth = node.Depth()>max_depth ? node.Depth() : max_depth;
      mesh.node_2d_list.push_back(node);
    }

    // Input of element-node connectivity
    for (std::size_t i=0; i<n_elements; i++) 
    {
      int id, n_vert;
      NODE_2D_SW *node_loc[vert_h];
      TRIANGLE_2D_SW element;

      g_coord >> id >> n_vert >> element.Node(0) >> element.Node(1) >> element.Node(2);
      element.ID()=i;
      if (id!=i+1)
      {
        std::cerr << "File mesh_container.cpp: Element ordering error at element: " << i << std::endl;
        std::exit(1);
      }
      g_coord.ignore(256, '\n');
      for (int j=0; j<vert_h; j++)
      {
        element.Node(j)--;
        node_loc[j]=&mesh.node_2d_list[element.Node(j)];
      }
      // Catch wrong orientation
      if (((node_loc[1]->Coord(0)-node_loc[0]->Coord(0))*
           (node_loc[2]->Coord(1)-node_loc[0]->Coord(1))-
           (node_loc[2]->Coord(0)-node_loc[0]->Coord(0))*
           (node_loc[1]->Coord(1)-node_loc[0]->Coord(1)))<0.)
      {
        int node_ind_tmp=element.Node(1);
        NODE_2D_SW *node_tmp=node_loc[1];

        element.Node(1)=element.Node(2);
        node_loc[1]=node_loc[2];
        element.Node(2)=node_ind_tmp;
        node_loc[2]=node_tmp;
      }
      element.ComputeReferenceMapping();
      element.SetActive();
      element.Update_basis_quad();
      triangle_2d_list.push_back(element);
    }

    // Input from "grid_connect" file
    // Edge-node and edge-element connectivity
    g_connect >> n_edges;
    g_connect.ignore(256, '\n');
    std::vector<EDGE_2D> edge_tmp_list(n_edges);
    for (std::size_t i=0; i<n_edges; i++)
    {
      int id;
      EDGE_2D_SW edge;

      g_connect >> id >> edge.Node(0) >> edge.Node(1) >> edge.Element(0) >> edge.Element(1);
      edge.ID()=i;
      if (id!=i+1)
      {
        std::cerr << "File mesh_container.cpp: Edge ordering error at edge: " << i << std::endl;
        std::exit(1);
      }
      g_connect.ignore(256, '\n');
      for (int j=0; j<2; j++)
      {
        edge.Node(j)--;
        edge.Element(j)--;
      }
      // Perform some simple checks
      for (int j=0; j<2; j++)
        if (edge.IsElement(j))
        {
          bool nd0=false, nd1=false;
          TRIANGLE_2D_SW& element_tmp=mesh.triangle_2d_list[edge.Element(j)];

          for (int k=0; k<vert_h; k++)
          {
            if (edge.Node(0)==element_tmp.Node(k))
              nd0=true;
            if (edge.Node(1)==element_tmp.Node(k))
              nd1=true;
          }
          if ((nd0==false) || (nd1==false))
          {
            std::cerr << "File mesh_container.cpp: Nodes from edge " << id << " missing on element " << 
              element_tmp.ID()+1 << std::endl;
            std::exit(1);
          }
        }
        edge_tmp_list[i]=edge;
    }

    // Element-edge connectivity
    for (std::size_t i=0; i<n_elements; i++)
    {
      int k, id, edge_loc[vert_h]={BAD_INDEX, BAD_INDEX, BAD_INDEX};
      TRIANGLE_2D_SW& element=mesh.triangle_2d_list[i];

      g_connect >> id >> element.Edge(0) >> element.Edge(1) >> element.Edge(2);
      if (id!=i+1)
      {
        std::cerr << "File mesh_container.cpp: Element ordering error at element: " << i << std::endl;
        std::exit(1);
      }
      g_connect.ignore(256, '\n');

      // Edge orientation check
      for (int j=0; j<vert_h; j++)
      {
        EDGE_2D& edge=edge_tmp_list[--element.Edge(j)];

        for (k=0; k<vert_h; k++)
          if ((edge.Node(0)!=element.Node(k)) && (edge.Node(1)!=element.Node(k)))
          {
            edge_loc[k]=element.Edge(j);
            break;
          }
          // Check edge-element connectivity and
          // order the nodes on the edge to conform to the ordering on the first element
          if (edge.Element(0)==element.ID())
          {
            element.ElementIndex(k)=0;
            edge.Node(0)=element.Node((k+1)%vert_h);
            edge.Node(1)=element.Node((k+2)%vert_h);
          }
          else if (edge.Element(1)==element.ID())
          {
            element.ElementIndex(k)=1;
            edge.Node(0)=element.Node((k+2)%vert_h);
            edge.Node(1)=element.Node((k+1)%vert_h);
          }
          else
          {
            std::cerr << "File mesh_container.cpp: Error in edge-element connectivity at element: " << id << std::endl;
            std::exit(1);
          }
      }
      for (int j=0; j<vert_h; j++) // Check element-edge connectivity
        if (edge_loc[j]==BAD_INDEX)
        {
          std::cerr << "File mesh_container.cpp: Error in element-edge connectivity at element: " << id << std::endl;
          std::exit(1);
        }
        else
          element.Edge(j)=edge_loc[j];  
    }

    {  // Input interior edges
      int n_interior;

      g_connect >> n_interior;
      g_connect.ignore(256, '\n');
      for (int i=0; i<n_interior; i++)
      {
        int id, ed;

        g_connect >> id >> ed;
        EDGE_INTERIOR edge(edge_tmp_list[ed-1]);
        edge_tmp_list[ed-1].Type()=EDGE_2D_SW::INTERIOR;
        g_connect.ignore(256, '\n');
        if (id!=i+1)
        {
          std::cerr << "File mesh_container.cpp: Interior edge ordering error at interior edge: " << i << std::endl;
          std::exit(1);
        }
        mesh.edge_2d_list.push_back(edge);
      }
    }

    {  // Input land edges
      int n_land;

      g_connect >> n_land;
      g_connect.ignore(256, '\n');
      for (int i=0; i<n_land; i++)
      {
        int id, ed;

        g_connect >> id >> ed;
        g_connect.ignore(256, '\n');
        EDGE_LAND edge(edge_tmp_list[ed-1]);
        edge_tmp_list[ed-1].Type()=EDGE_2D_SW::LAND;
        if (id!=i+1)
        {
          std::cerr << "File mesh_container.cpp: Land edge ordering error at land edge: " << i << std::endl;
          std::exit(1);
        }
        mesh.edge_2d_list.push_back(edge);
      }
    }

    {  // Input radiation edges
      int n_radiation;

      g_connect >> n_radiation;
      g_connect.ignore(256, '\n');
      for (int i=0; i<n_radiation; i++)
      {
        int id, ed;

        g_connect >> id >> ed;
        g_connect.ignore(256, '\n');
        EDGE_RADIATION edge(edge_tmp_list[ed-1]);
        edge_tmp_list[ed-1].Type()=EDGE_2D_SW::RADIATION;
        if (id!=i+1)
        {
          std::cerr << "File mesh_container.cpp: Radiation edge ordering error at radiation edge: " << i << std::endl;
          std::exit(1);
        }
        mesh.edge_2d_list.push_back(edge);
      }
    }

    {  // Input river edges
      int n_river;

      g_connect >> n_river;
      g_connect.ignore(256, '\n');
      for (int i=0; i<n_river; i++)
      {
        int id, ed;

        g_connect >> id >> ed;
        EDGE_RIVER edge(edge_tmp_list[ed-1]);
        edge_tmp_list[ed-1].Type()=EDGE_2D_SW::RIVER;
        g_connect >> edge.SurfaceElevation() >> edge.Velocity(0) >> edge.Velocity(1);
#ifdef BAROCLINIC_KEY
        real tmp;
        char buffer[300];
        g_connect.getline(buffer, 300);
        std::string strng(buffer);
        std::istringstream strin(strng);
        edge.Temperature()=problem_info.T_0, edge.Salinity()=problem_info.S_0;
        strin >> tmp;
        if (!strin.fail())
          edge.Temperature()=tmp;
        strin >> tmp;
        if (!strin.fail())
          edge.Salinity()=tmp;
#else
        g_connect.ignore(256, '\n');
#endif
        if (id!=i+1)
        {
          std::cerr << "File mesh_container.cpp: River edge ordering error at iver edge: " << i << std::endl;
          std::exit(1);
        }
        mesh.edge_2d_list.push_back(edge);
      }
    }

    {  // Input sea edges (special treatment)
      int n_sea;

      g_connect >> n_sea;
      g_connect.ignore(256, '\n');
      for (int i=0; i<n_sea; i++)
      {
        int id, ed;

        g_connect >> id >> ed;
        EDGE_SEA edge(edge_tmp_list[ed-1]);
        edge_tmp_list[ed-1].Type()=EDGE_2D_SW::SEA;
#ifdef BAROCLINIC_KEY
        real tmp;
        char buffer[300];
        g_connect.getline(buffer, 300);
        std::string strng(buffer);
        std::istringstream strin(strng);
        edge.Temperature()=problem_info.T_0, edge.Salinity()=problem_info.S_0;
        strin >> tmp;
        if (!strin.fail())
          edge.Temperature()=tmp;
        strin >> tmp;
        if (!strin.fail())
          edge.Salinity()=tmp;
#else
        g_connect.ignore(256, '\n');
#endif
        if (id!=i+1)
        {
          std::cerr << "File mesh_container.cpp: Sea edge ordering error at sea edge: " << i << std::endl;
          std::exit(1);
        }
        mesh.edge_2d_list.push_back(edge);
      }

      // Read in relative amplitude and phase
      for (int i=0; i<problem_info.n_bound_tp_freq; i++)
        for (int j=0; j<n_sea; j++)
        {
          EDGE_SEA& edge=mesh.edge_sea_list[j];

          g_connect >> edge.Amplitude(i) >> edge.Phase(i);
          g_connect.ignore(256, '\n');
          edge.Phase(i)*=problem_info.pi/180.;
          edge.ComputeTidalCoefficients(i);
        }
    }

#ifdef TEST_KEY
    {  // Input unknown edges
      int n_unknown=n_edges-mesh.edge_2d_list.size();

      for (int i=0; i<n_edges; i++)
        if (edge_tmp_list[i].Type()==EDGE_2D_SW::UNKNOWN)
        {
          EDGE_UNKNOWN edge(edge_tmp_list[i]);
          mesh.edge_2d_list.push_back(edge);
        }
    }
#endif

    // Complete the list of edges and check the edge type
    for (std::size_t i=0; i<n_edges; i++)
    {
      EDGE_2D_SW& edge=edge_2d_list[i];

#ifndef TEST_KEY
      if (edge.Type()==EDGE_2D_SW::UNKNOWN)
      {
        std::cerr << "File mesh_container.cpp: Type of edge " << i << " is unknown!" << std::endl;
        std::exit(1);
      }
#endif
      edge.SetActive();
      edge.ComputeReferenceMapping();
      edge.ComputeSurfaceElevation();
      edge.InitializeStorage();
      edge.Update_basis_quad();
      for (edge.SetIndex(0); edge.InRange(); edge++)
        edge.surface_elev_quad()=edge.SurfaceElevation(edge.Quad()->p(edge.GetIndex()));
    }

    edge_tmp_list.clear();
    g_coord.close();
    g_connect.close();
  }
  catch(std::exception& ex)
  {
    std::cerr << "File mesh_container.cpp: I/O error in configuration files" << ex.what() << 'n';
    std::exit(1);
  }
}

// Compute tidal forcing on elements and open sea edges
// Parameters: t - time
void MESH::ComputeTidalForcing(real t)
{
  // Compute tidal forcing on the elements
  if (problem_info.n_tp_freq>0)
  {
    std::vector<real> cost(problem_info.n_tp_freq), sint(problem_info.n_tp_freq);

    for (int i=0; i<problem_info.n_tp_freq; i++)
      cost[i]=problem_info.ramp_value*std::cos(problem_info.tp_param[i].amig*t), 
      sint[i]=problem_info.ramp_value*std::sin(problem_info.tp_param[i].amig*t);
    for (std::size_t k=0; k<mesh.triangle_2d_list.size(); k++)
    {
      TRIANGLE_2D_SW& element=mesh.triangle_2d_list[k];

      if (!element.IsActive())
        continue;
      element.TidalForcing(0)=element.TidalForcing(1)=0.;
      for (int i=0; i<problem_info.n_tp_freq; i++)
      {
        element.TidalForcing(0)+=element.TidalCoef(i, 0, 0)*cost[i]-element.TidalCoef(i, 1, 0)*sint[i];
        element.TidalForcing(1)+=element.TidalCoef(i, 0, 1)*cost[i]-element.TidalCoef(i, 1, 1)*sint[i];
      }
    }
  }

  // Compute tidal elevation on the open sea edges
  if (problem_info.n_bound_tp_freq>0)
  {
    std::vector<real> costb(problem_info.n_bound_tp_freq), sintb(problem_info.n_bound_tp_freq);

    for (int i=0; i<problem_info.n_bound_tp_freq; i++)
      costb[i]=problem_info.ramp_value*std::cos(problem_info.bound_tp_param[i].amig*t), 
      sintb[i]=problem_info.ramp_value*std::sin(problem_info.bound_tp_param[i].amig*t);
    for (std::size_t k=0; k<mesh.edge_sea_list.size(); k++)
    {
      EDGE_SEA& edge=mesh.edge_sea_list[k];

      if (!edge.IsActive())
        continue;
      edge.TidalElevation()=0.;
      for (int i=0; i<problem_info.n_bound_tp_freq; i++)
        edge.TidalElevation()+=edge.TidalCoef(i, 0)*costb[i]-edge.TidalCoef(i, 1)*sintb[i];
    }
  }
}

// Create 3D mesh
// Parameters: max_depth - maximum admissible bathymetric depth
void MESH::Create3DMesh(real max_depth)
{
  std::size_t n_node=node_2d_list.size(), n_edge=edge_2d_list.size(), 
    n_face=triangle_2d_list.size(), n_node_3d=0;
  long_real dz=max_depth/problem_info.n_layers;
  std::vector<std::pair<int, int> > node_index(n_node);
  std::vector<std::vector<int> > face_vert_index(n_edge);

  // Create 3D nodes
  for (std::size_t i=0; i<n_node; i++)
  {
    NODE_2D_SW& node_2d=node_2d_list[i];
    real depth=node_2d.Depth();
    int n_layers_loc=(int)std::ceil(depth/dz);

    if((n_layers_loc>1) && (depth-(n_layers_loc-1)*dz<problem_info.mesh_grading_bot*dz))
      n_layers_loc--;
    node_index[i].first=node_3d_list.size();
    node_index[i].second=n_layers_loc+1;
    for (int j=0; j<=n_layers_loc; j++)
    {
      NODE_3D_SW node_tmp(node_2d);
      node_tmp.ZCoordinate() = j<n_layers_loc ? -j*dz : -depth;
      node_tmp.ID()=n_node_3d++;
      node_3d_list.push_back(node_tmp);
    }
  }

  // Create elements and faces
  for (std::size_t i=0; i<n_face; i++)
  {
    int elem_count=problem_info.n_layers, node_in_edge_ind[vert_h][2];
    TRIANGLE_2D_SW *element_2d=&triangle_2d_list[i];
    int loc_ind[vert_h]={element_2d->ElementIndex(0), element_2d->ElementIndex(1), element_2d->ElementIndex(2)};
    EDGE_2D_SW *edge_2d[vert_h]={&edge_2d_list[element_2d->Edge(0)], 
      &edge_2d_list[element_2d->Edge(1)], 
      &edge_2d_list[element_2d->Edge(2)]};
    NODE_2D_SW *node_2d[vert_h]={&node_2d_list[element_2d->Node(0)], 
      &node_2d_list[element_2d->Node(1)], 
      &node_2d_list[element_2d->Node(2)]};
    std::pair<int, int> *node_3d_column[vert_h]={&node_index[node_2d[0]->ID()],
      &node_index[node_2d[1]->ID()],
      &node_index[node_2d[2]->ID()]};
    std::vector<int> *face_vert_column[vert_h]={&face_vert_index[edge_2d[0]->ID()],
      &face_vert_index[edge_2d[1]->ID()],
      &face_vert_index[edge_2d[2]->ID()]};

    // Determine local number of layers
    for (int j=0; j<vert_h; j++)
      elem_count = elem_count>node_3d_column[j]->second-1 ? node_3d_column[j]->second-1 : elem_count;

    // Set up various indexes and create vertical faces
    for (int j=0; j<vert_h; j++)
    {
      NODE_2D_SW *node_on_edge[2]={&node_2d_list[edge_2d[j]->Node(0)],
        &node_2d_list[edge_2d[j]->Node(1)]};

      for (int k=0; k<vert_h; k++)
      {
        if (node_on_edge[0]==node_2d[k])
          node_in_edge_ind[j][0]=k;
        if (node_on_edge[1]==node_2d[k])
          node_in_edge_ind[j][1]=k;
      }

      // Create vertical faces
      int neighbor_count=face_vert_column[j]->size();
      if (neighbor_count==elem_count) // Neighbor has same number of layers
        continue;
      if (neighbor_count==0) // Neighbor is not initialized yet or there is no neighbor
      {
        for (int k=0; k<elem_count; k++)
        {
          FACE_VERT_SW* face_vert_tmp=CreateFaceVert(*edge_2d[j]);
          face_vert_tmp->SetActive();
          face_vert_tmp->ID()=face_vert_list.size();
          face_vert_list.push_back(*face_vert_tmp);
          face_vert_column[j]->push_back(face_vert_tmp->ID());
          delete face_vert_tmp;
        }
        continue;
      }
      if (neighbor_count<elem_count) // Neighbor has fewer layers
      {
        int child=face_vert_list.size(), parent_ind=(*face_vert_column[j])[neighbor_count-1];

        face_vert_column[j]->pop_back();
        for (int k=neighbor_count-1; k<elem_count; k++)
        {
          FACE_VERT_SW *face_vert_tmp=CreateFaceVert(*edge_2d[j]);
          face_vert_tmp->SetActive();
          face_vert_tmp->Element(0)=face_vert_list[parent_ind].Element(0);
          face_vert_tmp->Element(1)=face_vert_list[parent_ind].Element(1);
          face_vert_tmp->Parent()=parent_ind;
          face_vert_tmp->ID()=face_vert_list.size();
          face_vert_list.push_back(*face_vert_tmp);
          face_vert_column[j]->push_back(face_vert_tmp->ID());
          delete face_vert_tmp;
        }
        face_vert_list[parent_ind].SetChildrenVert(elem_count-neighbor_count+1, child);
        face_vert_list[parent_ind].UnSetActive();
        face_vert_list[parent_ind].Status()|=FACE_VERT::VERT_SPLIT;
        continue;
      }
      if (neighbor_count>elem_count) // Neighbor has more layers
      {
        FACE_VERT_SW *face_vert_tmp=CreateFaceVert(*edge_2d[j]);
        face_vert_tmp->UnSetActive();
        face_vert_tmp->
          SetChildrenVert(neighbor_count-elem_count+1, (*face_vert_column[j])[elem_count-1]);
        face_vert_tmp->Status()|=FACE_VERT::VERT_SPLIT;
        face_vert_tmp->ID()=face_vert_list.size();
        for (int k=elem_count-1; k<neighbor_count; k++)
          face_vert_list[(*face_vert_column[j])[k]].Parent()=face_vert_tmp->ID();
        face_vert_list.push_back(*face_vert_tmp);
        (*face_vert_column[j])[elem_count-1]=face_vert_tmp->ID();
        face_vert_column[j]->erase(face_vert_column[j]->begin()+elem_count, 
          face_vert_column[j]->end());
        delete face_vert_tmp;
        continue;
      }
    }

    // Create 3D elements
    for (int j=0; j<elem_count; j++)
    {
      NODE_3D_SW *node_3d[2][vert_h]={{0, 0, 0}, {0, 0, 0}};
      FACE_VERT_SW *face_vert[vert_h]={0, 0, 0};
      PRISM_SW *prism;
      if  ((j==0) || (j==elem_count-1))
        {
          prism = new PRISM_NONPARALLEL(*element_2d);
          assert(prism!=0);
          if (j==0)
            prism->SetSurface();
          if (j==elem_count-1)
            prism->SetBottom();
//           if (j==0)
//             if (j==elem_count-1)
//               prism->Type()=PRISM_SW::SURFACE_AND_BOTTOM;
//             else
//               prism->Type()=PRISM_SW::SURFACE;
//           else
//             prism->Type()=PRISM_SW::BOTTOM;
        }
      else
        {
          prism = new PRISM_PARALLEL(*element_2d);
          assert(prism!=0);
//           prism->Type()=PRISM_SW::PARALLEL;
        }

      prism->ID()=prism_list.size();
      // Assign nodes to the bottom and top faces
      for (int k=0; k<vert_h; k++)
      {
        node_3d[0][k]=&node_3d_list[node_3d_column[k]->first+j];
        int ind = j==elem_count-1 ? 
          node_3d_column[k]->first+node_3d_column[k]->second-1 : 
        node_3d_column[k]->first+j+1;
        node_3d[1][k]=&node_3d_list[ind];
      }

      // Create the surface face in a column
      if(j==0)
      {
        FACE_HORIZ_SW *face_horiz;

        switch(problem_info.surface_bc)
        {
        case 0: face_horiz = new FACE_HORIZ_SURFACE(*element_2d);
          break;
        case 1:
        case 2: face_horiz = new FACE_HORIZ_SURFACE_STRESS(*element_2d);
          break;
        default: std::cerr << "File mesh_container.cpp: Attempting to create a surface face of unknown type!\n";
        }
        face_horiz->ID()=face_horiz_list.size();
        face_horiz->Element(0)=BAD_INDEX;
        face_horiz->Element(1)=prism->ID();
        for (int k=0; k<vert_h; k++)
          face_horiz->Node(k)=node_3d[0][k]->ID();
        face_horiz->SetActive();
        face_horiz_list.push_back(*face_horiz);
        for (int k=0; k<vert_h; k++)
          node_3d[0][k]->SetSurface();
        delete face_horiz;
      }

      // Assign the top face of the element
      face_horiz_list[face_horiz_list.size()-1].Element(1)=prism->ID();
      prism->TopFace()=face_horiz_list[face_horiz_list.size()-1].ID();

      // Create the bottom face of an element
      FACE_HORIZ_SW *face_horiz;
      if (j==(elem_count-1))
        switch(problem_info.bottom_bc)
      {
        case 1: face_horiz = new FACE_HORIZ_BOTTOM_NO_SLIP(*element_2d);
          break;
        case 2: face_horiz = new FACE_HORIZ_BOTTOM_NO_DRAG(*element_2d);
          break;
        case 3: face_horiz = new FACE_HORIZ_BOTTOM_QUAD_DRAG(*element_2d);
          break;
        case 4: face_horiz = new FACE_HORIZ_BOTTOM_LIN_DRAG(*element_2d);
          break;
#ifdef TEST_KEY
        case 5: face_horiz = new FACE_HORIZ_BOTTOM_TEST(*element_2d);
          break;
#endif
        default: std::cerr << "File mesh_container.cpp: Attempting to create a bottom face of unknown type!\n";
      }
      else
        face_horiz = new FACE_HORIZ_INTERIOR(*element_2d);
      assert(face_horiz!=0);
      face_horiz->ID()=face_horiz_list.size();
      face_horiz->Element(0)=prism->ID();
      prism->BottomFace()=face_horiz->ID();
      for (int k=0; k<vert_h; k++)
        face_horiz->Node(k)=node_3d[1][k]->ID();
      face_horiz->SetActive();
      face_horiz_list.push_back(*face_horiz);
      delete face_horiz;

      // Assign vertical faces
      for (int k=0; k<vert_h; k++)
      {
        NODE_3D_SW *node_3d_tmp[4]=
        {node_3d[0][node_in_edge_ind[k][0]], node_3d[1][node_in_edge_ind[k][0]], 
        node_3d[0][node_in_edge_ind[k][1]], node_3d[1][node_in_edge_ind[k][1]]};

        face_vert[k]=&face_vert_list[(*face_vert_column[k])[j]];
        // Set face-element index for the vertical face
        face_vert[k]->SetElementRecursive(loc_ind[k], prism->ID());
        if (j==0) // Tag vertical faces in the surface layer
          face_vert[k]->SetSurface();
        for (int l=0; l<4; l++)
          face_vert[k]->Node(l)=node_3d_tmp[l]->ID();
        prism->SideFace(k)=face_vert[k]->ID();
        //               face_vert[k]->ComputeReferenceMapping();
      }

      // Complete the geometry of the element
      prism->SetActive();
      prism_list.push_back(*prism);
      delete prism;
    }
  }

  // Remove temporary storage for nodes
  node_index.clear();

  // Remove temporary storage for vertical faces
  face_vert_index.clear();

  // Initialize static data
  EDGE_2D_SW::Update_static(var, var_rhs);
  TRIANGLE_2D_SW::Update_static(quadrature_index.GetTriangleQuad(problem_info.def_order), 
                                quadrature_index.GetTriangleNodeQuad(), (const BASIS**)basis_2d, var, var_rhs);
  FACE_HORIZ_SW::Update_static(var, var_flux, var_rhs);
  FACE_VERT_SW::Update_static(var, var_flux, var_rhs);
  PRISM_SW::Update_static(quadrature_index.GetPrismQuad(problem_info.def_order), 
                          (const BASIS**)basis_ref, var, var_flux, var_rhs);
  n_prism_active=prism_list.size();

  // Set quadrature on elements
  for (std::size_t i=0; i<prism_list.size(); i++)
    if (prism_list[i].IsActive())
    {
      prism_list[i].ComputeReferenceMapping();
      prism_list[i].InitializeStorage();
      prism_list[i].Update_basis_quad();
    }

  // Set quadrature on horizontal faces
  for (std::size_t i=0; i<face_horiz_list.size(); i++)
    if (face_horiz_list[i].IsActive())
    {
      face_horiz_list[i].ComputeReferenceMapping();
      face_horiz_list[i].InitializeStorage();
      face_horiz_list[i].Update_basis_quad();
    }

  // Set quadrature on vertical faces
  for (std::size_t i=0; i<face_vert_list.size(); i++)
    if (face_vert_list[i].IsActive())
    {
      face_vert_list[i].ComputeReferenceMapping();
      face_vert_list[i].InitializeStorage();
      face_vert_list[i].Update_basis_quad();
    }

  // Mesh consistency check
  for (std::size_t i=0; i<prism_list.size(); i++)
  {
    PRISM_SW& prism=prism_list[i];

    if (face_horiz_list[prism.TopFace()].Element(1)!=prism.ID())
    {
      std::cerr << "File mesh_container.cpp: Top face inconsistency at element: " << prism.ID() << std::endl;
      std::exit(1);
    }
    if (face_horiz_list[prism.BottomFace()].Element(0)!=prism.ID())
    {
      std::cerr << "File mesh_container.cpp: Bottom face inconsistency at element: " << prism.ID() << std::endl;
      std::exit(1);
    }
    for (int j=0; j<vert_h; j++)
      if (face_vert_list[prism.SideFace(j)].Element(prism.ElementIndex(j))!=prism.ID())
      {
        std::cerr << "File mesh_container.cpp: Vertical face inconsistency at element: " << prism.ID() << std::endl;
        std::exit(1);
      }
    for (int k=0; k<vert_h; k++)
      for (int j=0; j<4; j++)
      {
        int l, node_ind=face_vert_list[prism.SideFace(k)].Node(j);

        for (l=0; l<vert_h; l++)
          if (j%2==0)
          {
            if (node_ind==face_horiz_list[prism.TopFace()].Node(l))
              break;
          }
          else
          {
            if (node_ind==face_horiz_list[prism.BottomFace()].Node(l))
              break;
          }
        if (l==vert_h)
        {
          std::cerr << "File mesh_container.cpp: Node inconsistency at element: " << prism.ID() << std::endl;
          std::exit(1);
        }
      }
  }
}

// Update free surface elevation
// Parameters: t - time
void MESH::UpdateSurfaceElevation(real t)
{
  // Update surface elevation at the nodes
  for (std::size_t i=0; i<node_2d_list.size(); i++)
    node_2d_list[i].Xi()=0., node_2d_list[i].Scale()=0.;
  for (std::size_t i=0; i<triangle_2d_list.size(); i++)
    if (triangle_2d_list[i].IsActive())
      for (int j=0; j<vert_h; j++)
      {
        NODE_2D_SW& node=node_2d_list[triangle_2d_list[i].Node(j)];
        node.Xi()=
          (triangle_2d_list[i].Var_dof(XI)*triangle_2d_list[i].basis_func_node(XI, j)*
           triangle_2d_list[i].Mapping().GetScale()+node.Xi()*node.Scale())/
          (node.Scale()+triangle_2d_list[i].Mapping().GetScale());
        node.Scale()+=triangle_2d_list[i].Mapping().GetScale();
      }
#ifdef TEST_KEY
  for (std::size_t i=0; i<edge_unknown_list.size(); i++)
    if (edge_unknown_list[i].IsActive())
      for (int j=0; j<2; j++)
      {
        real surface_elevation;
        NODE_2D_SW& node=node_2d_list[edge_unknown_list[i].Node(j)];
        solution_analytic_elevation(t, node.Coord(), &surface_elevation);
        node.Xi()=surface_elevation;
      }
#endif
  // Compute elevation on 2D edges
  for (std::size_t i=0; i<edge_2d_list.size(); i++)
    if (edge_2d_list[i].IsActive())
    {
      real elev[2];
      EDGE_2D_SW& edge_sw=edge_2d_list[i];

      switch (edge_sw.Type())
      {
      case EDGE_2D_SW::RIVER:
        elev[0]=elev[1]=
          problem_info.ramp_value*((EDGE_RIVER&)edge_sw).SurfaceElevation();
        edge_sw.ComputeSurfaceElevation(elev);
        break;
      case EDGE_2D_SW::SEA:
        elev[0]=elev[1]=((EDGE_SEA&)edge_sw).TidalElevation();
        edge_sw.ComputeSurfaceElevation(elev);
        break;
      default: edge_sw.ComputeSurfaceElevation();
      }
      for (edge_sw.SetIndex(0); edge_sw.InRange(); edge_sw++)
      {
        const real *xi=edge_sw.Quad()->p(edge_sw.GetIndex());
        edge_sw.surface_elev_quad()=edge_sw.SurfaceElevation(xi);
      }
    }

  // Update vertical nodes in the surface layer
  for (std::size_t i=0; i<node_3d_list.size(); i++)
    if (node_3d_list[i].IsSurface())
      node_3d_list[i].ZCoordinate()=node_2d_list[node_3d_list[i].Node2D()].Xi();

  // Update elements and horizontal faces
  for (std::size_t i=0; i<face_horiz_list.size(); i++)
    if (face_horiz_list[i].IsSurface() && face_horiz_list[i].IsActive())
    {
      FACE_HORIZ_SW& face_horiz=face_horiz_list[i];
      PRISM_SW& prism=prism_list[face_horiz.Element(1)];

      triangle_2d_list[face_horiz.Element2D()].ComputeSurfaceElevation();
      face_horiz.UpdateSurfaceElevation();
      prism.UpdateSurfaceElevation();
      face_horiz_list[prism.BottomFace()].UnSetUpdated();
      for (int j=0; j<vert_h; j++)
        face_vert_list[prism.SideFace(j)].UnSetUpdatedRecursive();
    }

  // Update vertical faces
  for (std::size_t i=0; i<face_vert_list.size(); i++)
    if (face_vert_list[i].IsSurface() && face_vert_list[i].IsActive())
      face_vert_list[i].UpdateSurfaceElevation();
}

#ifdef ADAPTIVITY_H_2D_KEY
// Update mesh in case of adaptive mesh refinement for 2D variables
void MESH::UpdateMesh()
{
  std::size_t n_triangle=triangle_2d_list.size(), n_prism=prism_list.size();
  bool updated=false;

  for (int i=n_triangle-1; i>=0; i--) // Compute shock indicator
  {
    TRIANGLE_2D_SW& element_2d=triangle_2d_list[i];
    if (element_2d.IsActive())
    {
      if (element_2d.JumpBoundary(0)<1.e-10)
        continue;
      // 0 - inflow, 1 - outflow
      real jump=std::abs(element_2d.Jump(0))/element_2d.JumpBoundary(0),
        norm_factor=std::pow((long_real)element_2d.Mapping().GetDiameter(), 
                             (long_real)(0.5*(element_2d.Order()+1.))), 
        shock_indicator=jump/norm_factor;
      if (shock_indicator>problem_info.detector_sensitivity) // Set refinement switch
      {
        if (element_2d.Parent()!=BAD_INDEX)
          mesh.triangle_2d_list[element_2d.Parent()].UnSetToBeUnRefined();
        if (element_2d.RefinementLevel()<problem_info.max_refinement)
        {
          element_2d.SetToBeRefined();
          element_2d.Refine();
          updated=true;
        }
      }
    }
    if (element_2d.IsToBeUnRefined())
    {
      element_2d.UnRefine();
      updated=true;
    }
  }
  // Check if any further action needed
  if (!updated)
    return;
  // Update quantities stored at quadrature points
  for (std::size_t j=0; j<mesh.edge_2d_list.size(); j++)
    if (mesh.edge_2d_list[j].IsActive() && !mesh.edge_2d_list[j].IsUpdated())
      mesh.edge_2d_list[j].Update_basis_quad();
  for (std::size_t i=0; i<n_prism; i++)
  {
    PRISM_SW& prism=prism_list[i];
    if (prism.IsActive() && triangle_2d_list[prism.Element2D()].IsToBeRefined())
    {
      prism.Refine();
      status|=UPDATED;
    }
    if (triangle_2d_list[prism.Element2D()].IsToBeUnRefined())
    {
      prism.UnRefine();
      status|=UPDATED;
    }
  }
  // Reserve storage for degrees of freedom on new elements
  Initialize_DOF();

  // Re-assign recording stations
  elev_stations.AssignStations(triangle_2d_list);
  vel_stations.AssignStations(triangle_2d_list);
}
#endif

#ifdef ADAPTIVITY_H_3D_KEY
// Update mesh in case of adaptive mesh refinement for 3D variables
void MESH::UpdateMesh()
{
  std::size_t n_triangle=triangle_2d_list.size(), n_prism=prism_list.size();
  bool updated=false;

  for (std::size_t i=0; i<n_prism; i++)// Compute shock indicator
  {
    PRISM_SW& prism=prism_list[i];
    TRIANGLE_2D_SW& element_2d=triangle_2d_list[prism.Element2D()];
    if (prism.IsActive())
    {
      if (prism.JumpBoundary(0)<1.e-10)
        continue;
      real jump=std::abs(prism.Jump(0))/prism.JumpBoundary(0),  // 0 - inflow, 1 - outflow
        norm_factor=std::pow((long_real)prism.Mapping().GetDiameter(),
                             (long_real)(0.5*(prism.Order()+1.))), 
        shock_indicator=jump/norm_factor;
      if (shock_indicator>problem_info.detector_sensitivity)
      { // Set refinement switch
        if (element_2d.Parent()!=BAD_INDEX)
          triangle_2d_list[element_2d.Parent()].UnSetToBeUnRefined();
        if (element_2d.RefinementLevel()<problem_info.max_refinement)
          element_2d.SetToBeRefined();
      }
    }
  }
  for (int i=n_triangle-1; i>=0; i--)
  {
    TRIANGLE_2D_SW& element_2d=triangle_2d_list[i];
    if (element_2d.IsActive() && element_2d.IsToBeRefined())
    {
      element_2d.Refine();
      updated=true;
    }
    if (element_2d.IsToBeUnRefined())
    {
      element_2d.UnRefine();
      updated=true;
    }
  }
  // Check if any further action needed
  if (!updated)
    return;
  // Update quantities stored at quadrature points
  for (std::size_t j=0; j<mesh.edge_2d_list.size(); j++)
    if (mesh.edge_2d_list[j].IsActive() && !mesh.edge_2d_list[j].IsUpdated())
      mesh.edge_2d_list[j].Update_basis_quad();
  for (std::size_t i=0; i<n_prism; i++)
  {
    PRISM_SW& prism=prism_list[i];
    if (prism.IsActive() && triangle_2d_list[prism.Element2D()].IsToBeRefined())
    {
      prism.Refine();
      status|=UPDATED;
    }
    if (triangle_2d_list[prism.Element2D()].IsToBeUnRefined())
    {
      prism.UnRefine();
      status|=UPDATED;
    }
  }
  // Reserve storage for degrees of freedom on new elements
  Initialize_DOF();

  // Re-assign recording stations
  elev_stations.AssignStations(triangle_2d_list);
  vel_stations.AssignStations(triangle_2d_list);
}
#endif

// Output in various formats
// Parameters: t - time; output_ind - type of required output
void MESH::Output(real t, int output_ind)
{
#ifdef TEST_KEY
  const int n_outputs=20;
  static long_real start_time=86400.*problem_info.start_time, output_time=start_time,
    end_time=start_time+86400.*problem_info.simul_length, 
    output_step=(end_time-start_time)/n_outputs, total_error[d_var_3d]={0.};
#endif
  const int norm_type=1;

  switch(output_ind)
  {
  case 0: // Output at the start time
    std::cout << "SIMULATION NAME: " << problem_info.run_description 
         << "   OUTPUT NAME: " << problem_info.output_file << std::endl
         << "Order of approximating space: " << problem_info.def_order 
         << "     Runge-Kutta scheme order: " << problem_info.runge_kutta_method << std::endl
         << "Number of surface elements: " << triangle_2d_list.size()   
         << "    Number of vertical layers: " << problem_info.n_layers << std::endl
         << "Simulation length (days): " << problem_info.simul_length 
         << "        Time step (seconds): " << problem_info.dt << std::endl << std::endl;
    break;

  case 1: // Output at the end of the simulation
    {
      TecplotOutputConforming(t);
      // Output recording stations
      elev_stations.Output(t);
      vel_stations.Output(t);

#ifdef TEST_KEY
      const int n_outputs=20;
      real error[d_var_3d]={0.}, error_elevation;

      // Update variable information at quadrature points
      for (std::size_t i=0; i<triangle_2d_list.size(); i++)
        if (triangle_2d_list[i].IsActive())
          triangle_2d_list[i].Update_variables_quad(t);
      global_error(t, solution_analytic, norm_type, error);
      error_elevation=elevation_error(t, solution_analytic_elevation, norm_type);
      std::cout << "Completed:" << std::setprecision(3) << 
        100.*(t-start_time)/(end_time-start_time) << "%," << " Elevation change="  
        << std::setprecision(8) << elevation_change(problem_info.dt, norm_type) 
        << " Elevation error="  << error_elevation << std::endl;
      for (int i=0; i<d_var_3d; i++)
        std::cout << " err(" << VAR_NAME[VAR_3D_INDEX[i]] << ")=" << error[i];
      std::cout << std::endl;
      if (norm_type==1)
        for (int i=0; i<d_var_3d; i++)
          total_error[i]=std::sqrt(total_error[i]);
      std::cout << std::endl << "Finish: ";
      for (int i=0; i<d_var_3d; i++)
        std::cout << " err(" << VAR_NAME[VAR_3D_INDEX[i]] << ")=" << total_error[i];
      std::cout << std::endl;
#endif
      break;
    }

  case 2: // Output error of the approximation and the state of the system at time t
    {
      TecplotOutputConforming(t);

      // Output recording stations
      elev_stations.Output(t);
      vel_stations.Output(t);

#ifdef TEST_KEY
      const int n_outputs=20;
      if (t>=output_time)
      {
        real error[d_var_3d]={0.}, error_elevation;

        //       ErrorOutput(t, solution_analytic);
        global_error(t, solution_analytic, norm_type, error);
        error_elevation=elevation_error(t, solution_analytic_elevation, norm_type);
        std::cout << "Completed:" << std::setprecision(3) << 
          100.*(t-start_time)/(end_time-start_time) << "%," << " Elevation change="  
          << std::setprecision(8) << elevation_change(problem_info.dt, norm_type) 
          << " Elevation error="  << error_elevation << std::endl;
        for (int i=0; i<d_var_3d; i++)
          std::cout << " err(" << VAR_NAME[VAR_3D_INDEX[i]] << ")=" << error[i];
        std::cout << std::endl;
        output_time+=output_step;
        if (t>start_time)
          switch (norm_type)
        {
          case 0:
            for (int i=0; i<d_var_3d; i++)
              total_error[i] = error[i]>total_error[i] ? error[i] : total_error[i];
            break;
          case 1:
            for (int i=0; i<d_var_3d; i++)
              total_error[i]+=error[i]*error[i]*problem_info.dt;
            break;
        }
      }
#endif
      break;
    }

  case 3: // Output change in the values of the state variables
#ifndef TEST_KEY
    if (output_counter-- == 0)
    {
      real dof_change[d_var_3d]={0.};

      output_counter=9;
      std::cout << std::setprecision(6) << "t=" << t << " sec (" << std::setprecision(8) << t/86400. 
           << " days)" 
#if defined(ADAPTIVITY_H_2D_KEY) || defined(ADAPTIVITY_H_3D_KEY)
           << " Number of 2D elements=" << triangle_2d_list.size()
#endif
           << "  Change(H)=" << elevation_change(problem_info.dt, norm_type);
#ifdef SUPERCRITICAL_KEY
      std::cout << "  Error(H)=" << elevation_error(t, solution_analytic_elevation, norm_type);
#endif
      std::cout << std::endl;
      global_change(problem_info.dt, norm_type, dof_change);
      for (int i=0; i<d_var_3d_dt; i++)
        std::cout << "  Change(" << VAR_NAME[VAR_3D_DT_INDEX[i]] << ")=" << dof_change[i];
      std::cout << std::endl;
#ifdef MASS_CONSERVATION_KEY
      std::cout << "  Tracer mass conservation (relative) error=" << tracer_mass_error(t) << std::endl;
#endif
      if (dof_change[VAR_3D_DT_INDEX_INV[U]]!=dof_change[VAR_3D_DT_INDEX_INV[U]])
        std::exit(1);
    }
#endif
    break;

  default: std::cerr << "File mesh_container.cpp: Wrong output choice parameter!" << std::endl;
    std::exit(1);
  }
}

// Project values of 3D unknowns to nodes for Tecplot output
// Parameters: n_nodes - number of nodes, lvalue; n_elements - number of elements, lvalue;
void MESH::ProjectToNodes(int& n_nodes, int& n_elements)
{
  n_nodes=n_elements=0;
  // Zero the data fields
  for (std::size_t i=0; i<node_3d_list.size(); i++)
  {
    node_3d_list[i].Scale()=0;
    for (int j=0; j<d_var_3d; j++)
      node_3d_list[i].Var(VAR_3D_INDEX[j])=0.;
  }

  // Compute values of variables at grid nodes
  for (std::size_t i=0; i<prism_list.size(); i++)
    if (prism_list[i].IsActive())
    {
      PRISM_SW& prism=prism_list[i];
      int node_ind[2*vert_h]={face_horiz_list[prism.TopFace()].Node(0),
        face_horiz_list[prism.TopFace()].Node(1),
        face_horiz_list[prism.TopFace()].Node(2),
        face_horiz_list[prism.BottomFace()].Node(0),
        face_horiz_list[prism.BottomFace()].Node(1),
        face_horiz_list[prism.BottomFace()].Node(2)};

      for (int j=0; j<2*vert_h; j++)
      {
        func_type basis_func;
        NODE_3D_SW& node=node_3d_list[node_ind[j]];

        node.Scale()+=prism.Mapping().GetScale();
        for (int k=0; k<d_var_3d; k++)
          node.Var(VAR_3D_INDEX[k])+=prism.Var_dof(VAR_3D_INDEX[k])*
            prism.basis_func_node(VAR_3D_INDEX[k], j, basis_func)*prism.Mapping().GetScale();
      }
      n_elements++;
    }
    // Count projected nodes
    for (std::size_t i=0; i<node_3d_list.size(); i++)
      if (node_3d_list[i].Scale()>0.)
      {
        n_nodes++;
        for (int j=0; j<d_var_3d; j++)
          node_3d_list[i].Var(VAR_3D_INDEX[j])/=node_3d_list[i].Scale();
      }
}

// Output in Tecplot format
// Parameters: t - time
void MESH::TecplotOutput(real t)
{
  int n_elements=mesh.n_prism_active, n_nodes=2*vert_h*n_elements;
  double t_days=t/86400.;

  if (!problem_info.elev_global)
    return;
  // Truncate if opening for the first time, otherwise append
  if (iteration_counter==0)
  {
    std::ofstream tec_tmp(problem_info.tecplot_out, std::ofstream::trunc);
    tec_tmp.close();
  }

  if ((iteration_counter==0) || (iteration_counter%problem_info.elev_global_freq==0 &&
    (t_days-problem_info.elev_global_start)*(problem_info.elev_global_end-t_days)>=0.))
  {
    std::ofstream tec_out(problem_info.tecplot_out, std::ios::in | std::ios::out | std::ios::ate);
    if (!tec_out.is_open())
    {
      std::cerr << "File mesh_container.cpp: Cannot open Tecplot output file!\n";
      std::exit(1);
    }
    if (tec_out.tellp()>tecplot_output_position)
      tec_out.seekp(tecplot_output_position, std::ios::beg);
    // Check if the file size exceeds 2GB 
    if (tec_out.tellp()>2.e9)
    {
      std::cerr << "File mesh_container.cpp: Output file size exceeds 2GB!\n";
      std::exit(1);
    }

    if (iteration_counter==0) // Output file header if opening for the first time
    {
      tec_out << "TITLE = \"" << problem_info.run_description << "\"" << std::endl
        << "VARIABLES = \"X\", \"Y\", \"Z\"";
      for (int i=0; i<d_var_3d; i++)
        tec_out << ", \"" << VAR_NAME[VAR_3D_INDEX[i]] << "\"";
      tec_out << std::endl;
    }
    // Output zone header
    tec_out << "\nZONE F=FEPOINT, ET=BRICK, T=\"Time=" << std::setprecision(10) << t 
      << " sec\", N=" << n_nodes << ", E=" << n_elements << std::endl;
    // Take connectivity from the previous zone unless opening for the first time
    if ((iteration_counter!=0) && !IsUpdated())
      tec_out << "D=(FECONNECT)" << std::endl;
    // Output values of unknowns at the nodes
    for (std::size_t i=0; i<prism_list.size(); i++)
      if (prism_list[i].IsActive())
      {
        PRISM_SW& prism=prism_list[i];
        int node_ind[2*vert_h]={face_horiz_list[prism.TopFace()].Node(0),
                                face_horiz_list[prism.TopFace()].Node(1),
                                face_horiz_list[prism.TopFace()].Node(2),
                                face_horiz_list[prism.BottomFace()].Node(0),
                                face_horiz_list[prism.BottomFace()].Node(1),
                                face_horiz_list[prism.BottomFace()].Node(2)};

        for (int j=0; j<2*vert_h; j++)
        {
          func_type basis_func;
          NODE_3D_SW& node=node_3d_list[node_ind[j]];

          tec_out << "   " << std::setprecision(10) << node.Coord(0) << "   " 
                  << node.Coord(1) << "   " << node.Coord(2);
          for (int k=0; k<d_var_3d; k++)
            tec_out << "  " << prism.Var_dof(VAR_3D_INDEX[k])*
              prism.basis_func_node(VAR_3D_INDEX[k], j, basis_func);
          tec_out << std::endl;
        }
      }
    // Output connectivity if the function is called for the first time
    if ((iteration_counter==0) || IsUpdated())
      for (std::size_t i=0, node_id=0; i<prism_list.size(); i++)
        if (prism_list[i].IsActive())
        {
          for (int k=0; k<2; k++)
          {
            for (int j=0; j<vert_h; j++)
              tec_out << "  " << ++node_id;
            tec_out << "  " << node_id;
          }
          tec_out << std::endl;
        }
    tecplot_output_position=tec_out.tellp();
    tec_out.close();
    status&=~UPDATED;
  }
}

// Output in Tecplot format (for conforming grids)
// Parameters: t - time
void MESH::TecplotOutputConforming(real t)
{
  int n_nodes, n_elements;
  double t_days=t/86400.;

  if (!problem_info.elev_global)
    return;
  // Truncate if opening for the first time, otherwise append
  if (iteration_counter==0)
  {
    std::ofstream tec_tmp(problem_info.tecplot_out, std::ofstream::trunc);
    tec_tmp.close();
  }

  if ((iteration_counter==0) || (iteration_counter%problem_info.elev_global_freq==0 &&
    (t_days-problem_info.elev_global_start)*(problem_info.elev_global_end-t_days)>=0.))
  {
    std::ofstream tec_out(problem_info.tecplot_out, std::ios::in | std::ios::out | std::ios::ate);
    if (!tec_out.is_open())
    {
      std::cerr << "File mesh_container.cpp: Cannot open Tecplot output file!\n";
      std::exit(1);
    }
    if (tec_out.tellp()>tecplot_output_position)
      tec_out.seekp(tecplot_output_position, std::ios::beg);
    // Check if the file size exceeds 2GB 
    if (tec_out.tellp()>2.e9)
    {
      std::cerr << "File mesh_container.cpp: Output file size exceeds 2GB!\n";
      std::exit(1);
    }

    if (iteration_counter==0) // Output file header if opening for the first time
    {
      tec_out << "TITLE = \"" << problem_info.run_description << "\"" << std::endl
        << "VARIABLES = \"X\", \"Y\", \"Z\"";
      for (int i=0; i<d_var_3d; i++)
        tec_out << ", \"" << VAR_NAME[VAR_3D_INDEX[i]] << "\"";
      tec_out << std::endl;
    }
    ProjectToNodes(n_nodes, n_elements); // Project values of the variables to nodes
    // Output zone header
    tec_out << "\nZONE F=FEPOINT, ET=BRICK, T=\"Time=" << std::setprecision(10) << t 
      << " sec\", N=" << n_nodes << ", E=" << n_elements << std::endl;
    // Take connectivity from the previous zone unless opening for the first time
    if ((iteration_counter!=0) && !IsUpdated())
      tec_out << "D=(FECONNECT)" << std::endl;
    // Output values of unknowns at the nodes
    for (std::size_t i=0, tmp_id=1; i<node_3d_list.size(); i++)
      if (node_3d_list[i].Scale()>0.)
      {
        NODE_3D_SW& node=node_3d_list[i];

        node.TMP_ID()=tmp_id++;
        tec_out << "   " << std::setprecision(10) << node.Coord(0) << "   " 
          << node.Coord(1) << "   " << node.Coord(2);
        for (int j=0; j<d_var_3d; j++)
          tec_out << "  " << node.Var(VAR_3D_INDEX[j]);
        tec_out << std::endl;
      }
      // Output connectivity if the function is called for the first time
      if ((iteration_counter==0) || IsUpdated())
        for (std::size_t i=0; i<prism_list.size(); i++)
          if (prism_list[i].IsActive())
          {
            PRISM_SW& prism=prism_list[i];
            FACE_HORIZ_SW &top=face_horiz_list[prism.TopFace()],
              &bot=face_horiz_list[prism.BottomFace()];

            tec_out << "  " << node_3d_list[top.Node(0)].TMP_ID()
              << "  " << node_3d_list[top.Node(1)].TMP_ID()
              << "  " << node_3d_list[top.Node(2)].TMP_ID()
              << "  " << node_3d_list[top.Node(2)].TMP_ID()
              << "  " << node_3d_list[bot.Node(0)].TMP_ID()
              << "  " << node_3d_list[bot.Node(1)].TMP_ID()
              << "  " << node_3d_list[bot.Node(2)].TMP_ID()
              << "  " << node_3d_list[bot.Node(2)].TMP_ID() << std::endl;
          }
          tecplot_output_position=tec_out.tellp();
          tec_out.close();
          status&=~UPDATED;
  }
}

// Hot start output
// Parameters: t - time
void MESH::HotOutput(long_real t)
{
//   char hot_file[file_name];
  std::size_t list_length;
  std::string hot_file(problem_info.output_file);

  // Construct the hot output file name
  if(problem_info.hot_output==2)
  {
    std::ostringstream strout;

    strout << iteration_counter;
    hot_file+="_iteration_";
    hot_file+=strout.str();
  }
  hot_file+=".hot";
  std::ofstream hot_out(hot_file.c_str(), std::ofstream::trunc | std::ofstream::binary);

//   std::strcpy(hot_file, problem_info.output_file);
//   switch(problem_info.hot_output)
//   {
//   case 1: std::strcat(hot_file, ".hot");
//     break;
//   case 2: std::strcat(hot_file, "_iteration_");
//     std::sprintf(hot_file+std::strlen(hot_file), "%d", iteration_counter);
//     std::strcat(hot_file, ".hot");
//     break;
//   }
//   std::ofstream hot_out(hot_file, std::ofstream::trunc | std::ofstream::binary);
  if (!hot_out)
  {
    std::cerr << "File mesh_container.cpp: Cannot open hot output file!" << std::endl;
    std::exit(1);
  }

  // Output global parameters
  hot_out.write((char*)&t, sizeof(t));
  hot_out.write((char*)&status, sizeof(status));
  hot_out.write((char*)&iteration_counter, sizeof(iteration_counter));
  hot_out.write((char*)&update_counter, sizeof(update_counter));
  hot_out.write((char*)&output_counter, sizeof(output_counter));
  hot_out.write((char*)&tecplot_output_position, sizeof(tecplot_output_position));

  // Output 2D nodes
  list_length=node_2d_list.size();
  hot_out.write((char*)&list_length, sizeof(list_length));
  for (std::size_t i=0; i<list_length; i++)
    hot_out << node_2d_list[i];

  // Output 2D edges
  list_length=edge_interior_list.size();
  hot_out.write((char*)&list_length, sizeof(list_length));
  for (std::size_t i=0; i<list_length; i++)
    hot_out << edge_interior_list[i];
  list_length=edge_land_list.size();
  hot_out.write((char*)&list_length, sizeof(list_length));
  for (std::size_t i=0; i<list_length; i++)
    hot_out << edge_land_list[i];
  list_length=edge_river_list.size();
  hot_out.write((char*)&list_length, sizeof(list_length));
  for (std::size_t i=0; i<list_length; i++)
    hot_out << edge_river_list[i];
  list_length=edge_sea_list.size();
  hot_out.write((char*)&list_length, sizeof(list_length));
  for (std::size_t i=0; i<list_length; i++)
    hot_out << edge_sea_list[i];
  list_length=edge_radiation_list.size();
  hot_out.write((char*)&list_length, sizeof(list_length));
  for (std::size_t i=0; i<list_length; i++)
    hot_out << edge_radiation_list[i];
#ifdef TEST_KEY
  list_length=edge_unknown_list.size();
  hot_out.write((char*)&list_length, sizeof(list_length));
  for (std::size_t i=0; i<list_length; i++)
    hot_out << edge_unknown_list[i];
#endif

  // Output 2D elements
  list_length=triangle_2d_list.size();
  hot_out.write((char*)&list_length, sizeof(list_length));
  for (std::size_t i=0; i<list_length; i++)
    hot_out << triangle_2d_list[i];

  // Output 3D nodes
  list_length=node_3d_list.size();
  hot_out.write((char*)&list_length, sizeof(list_length));
  for (std::size_t i=0; i<list_length; i++)
    hot_out << node_3d_list[i];

  // Output horizontal faces
  list_length=face_horiz_surface_list.size();
  hot_out.write((char*)&list_length, sizeof(list_length));
  for (std::size_t i=0; i<list_length; i++)
    hot_out << face_horiz_surface_list[i];
  list_length=face_horiz_surface_stress_list.size();
  hot_out.write((char*)&list_length, sizeof(list_length));
  for (std::size_t i=0; i<list_length; i++)
    hot_out << face_horiz_surface_stress_list[i];
  list_length=face_horiz_interior_list.size();
  hot_out.write((char*)&list_length, sizeof(list_length));
  for (std::size_t i=0; i<list_length; i++)
    hot_out << face_horiz_interior_list[i];
  list_length=face_horiz_bottom_no_slip_list.size();
  hot_out.write((char*)&list_length, sizeof(list_length));
  for (std::size_t i=0; i<list_length; i++)
    hot_out << face_horiz_bottom_no_slip_list[i];
  list_length=face_horiz_bottom_no_drag_list.size();
  hot_out.write((char*)&list_length, sizeof(list_length));
  for (std::size_t i=0; i<list_length; i++)
    hot_out << face_horiz_bottom_no_drag_list[i];
  list_length=face_horiz_bottom_quad_drag_list.size();
  hot_out.write((char*)&list_length, sizeof(list_length));
  for (std::size_t i=0; i<list_length; i++)
    hot_out << face_horiz_bottom_quad_drag_list[i];
  list_length=face_horiz_bottom_lin_drag_list.size();
  hot_out.write((char*)&list_length, sizeof(list_length));
  for (std::size_t i=0; i<list_length; i++)
    hot_out << face_horiz_bottom_lin_drag_list[i];
#ifdef TEST_KEY
  list_length=face_horiz_bottom_test_list.size();
  hot_out.write((char*)&list_length, sizeof(list_length));
  for (std::size_t i=0; i<list_length; i++)
    hot_out << face_horiz_bottom_test_list[i];
#endif

  // Output vertical faces
  list_length=face_vert_interior_list.size();
  hot_out.write((char*)&list_length, sizeof(list_length));
  for (std::size_t i=0; i<list_length; i++)
    hot_out << face_vert_interior_list[i];
  list_length=face_vert_land_list.size();
  hot_out.write((char*)&list_length, sizeof(list_length));
  for (std::size_t i=0; i<list_length; i++)
    hot_out << face_vert_land_list[i];
  list_length=face_vert_river_list.size();
  hot_out.write((char*)&list_length, sizeof(list_length));
  for (std::size_t i=0; i<list_length; i++)
    hot_out << face_vert_river_list[i];
  list_length=face_vert_sea_list.size();
  hot_out.write((char*)&list_length, sizeof(list_length));
  for (std::size_t i=0; i<list_length; i++)
    hot_out << face_vert_sea_list[i];
  list_length=face_vert_radiation_list.size();
  hot_out.write((char*)&list_length, sizeof(list_length));
  for (std::size_t i=0; i<list_length; i++)
    hot_out << face_vert_radiation_list[i];
#ifdef TEST_KEY
  list_length=face_vert_unknown_list.size();
  hot_out.write((char*)&list_length, sizeof(list_length));
  for (std::size_t i=0; i<list_length; i++)
    hot_out << face_vert_unknown_list[i];
#endif
  // Output vertical split child index storage for vertical faces
  FACE_VERT::OutputChildVert(hot_out);
  

  // Output 3D elements
  list_length=prism_parallel_list.size();
  hot_out.write((char*)&list_length, sizeof(list_length));
  for (std::size_t i=0; i<list_length; i++)
    hot_out << prism_parallel_list[i];
  list_length=prism_nonparallel_list.size();
  hot_out.write((char*)&list_length, sizeof(list_length));
  for (std::size_t i=0; i<list_length; i++)
    hot_out << prism_nonparallel_list[i];

  // Output primary varibles
  for (std::size_t i=0; i<d_var; i++)
    hot_out << var[i];

  // Output recording stations info
  hot_out << elev_stations;
  hot_out << vel_stations;
  hot_out.close();

  std::cout << "Finished outputting to hot start file: " << hot_file << std::endl;
}


// Hot start input
// Parameters: t - time, lvalue
void MESH::HotInput(long_real& t)
{
  // Initialize static data
  EDGE_2D_SW::Update_static(var, var_rhs);
  TRIANGLE_2D_SW::Update_static(quadrature_index.GetTriangleQuad(problem_info.def_order), 
                                quadrature_index.GetTriangleNodeQuad(), (const BASIS**)basis_2d, var, var_rhs);
  FACE_HORIZ_SW::Update_static(var, var_flux, var_rhs);
  FACE_VERT_SW::Update_static(var, var_flux, var_rhs);
  PRISM_SW::Update_static(quadrature_index.GetPrismQuad(problem_info.def_order), 
                          (const BASIS**)basis_ref, var, var_flux, var_rhs);

  std::ifstream hot_in(problem_info.hot_input_file, std::ifstream::binary);
  if (!hot_in)
  {
    std::cerr << "File mesh_container.cpp: Cannot open hot input file!" << std::endl;
    std::exit(1);
  }

  std::cout << "Using input from hot start file: " << problem_info.hot_input_file << std::endl;
  try
  {
    std::size_t list_length;

    // Input global parameters
    hot_in.read((char*)&t, sizeof(t));
    hot_in.read((char*)&status, sizeof(status));
    hot_in.read((char*)&iteration_counter, sizeof(iteration_counter));
    hot_in.read((char*)&update_counter, sizeof(update_counter));
    hot_in.read((char*)&output_counter, sizeof(output_counter));
    if (output_counter-- == 0)
      output_counter=9;
    hot_in.read((char*)&tecplot_output_position, sizeof(tecplot_output_position));

    // Input 2D nodes
    hot_in.read((char*)&list_length, sizeof(list_length));
    for (std::size_t i=0; i<list_length; i++)
    {
      NODE_2D_SW node(hot_in);
      node_2d_list.push_back(node);
    }

    // Input 2D edges
    hot_in.read((char*)&list_length, sizeof(list_length));
    for (std::size_t i=0; i<list_length; i++)
    {
      EDGE_INTERIOR edge(hot_in);
      edge_2d_list.push_back(edge);
    }
    hot_in.read((char*)&list_length, sizeof(list_length));
    for (std::size_t i=0; i<list_length; i++)
    {
      EDGE_LAND edge(hot_in);
      edge_2d_list.push_back(edge);
    }
    hot_in.read((char*)&list_length, sizeof(list_length));
    for (std::size_t i=0; i<list_length; i++)
    {
      EDGE_RIVER edge(hot_in);
      edge_2d_list.push_back(edge);
    }
    hot_in.read((char*)&list_length, sizeof(list_length));
    for (std::size_t i=0; i<list_length; i++)
    {
      EDGE_SEA edge(hot_in);
      edge_2d_list.push_back(edge);
    }
    hot_in.read((char*)&list_length, sizeof(list_length));
    for (std::size_t i=0; i<list_length; i++)
    {
      EDGE_RADIATION edge(hot_in);
      edge_2d_list.push_back(edge);
    }
#ifdef TEST_KEY
    hot_in.read((char*)&list_length, sizeof(list_length));
    for (std::size_t i=0; i<list_length; i++)
    {
      EDGE_UNKNOWN edge(hot_in);
      edge_2d_list.push_back(edge);
    }
#endif

    // Input 2D elements
    hot_in.read((char*)&list_length, sizeof(list_length));
    for (std::size_t i=0; i<list_length; i++)
    {
      TRIANGLE_2D_SW element(hot_in);
      triangle_2d_list.push_back(element);
    }

    // Input 3D nodes
    hot_in.read((char*)&list_length, sizeof(list_length));
    for (std::size_t i=0; i<list_length; i++)
    {
      NODE_3D_SW node(hot_in);
      node_3d_list.push_back(node);
    }

    // Input horizontal faces
    hot_in.read((char*)&list_length, sizeof(list_length));
    for (std::size_t i=0; i<list_length; i++)
    {
      FACE_HORIZ_SURFACE face(hot_in);
      face_horiz_list.push_back(face);
    }
    hot_in.read((char*)&list_length, sizeof(list_length));
    for (std::size_t i=0; i<list_length; i++)
    {
      FACE_HORIZ_SURFACE_STRESS face(hot_in);
      face_horiz_list.push_back(face);
    }
    hot_in.read((char*)&list_length, sizeof(list_length));
    for (std::size_t i=0; i<list_length; i++)
    {
      FACE_HORIZ_INTERIOR face(hot_in);
      face_horiz_list.push_back(face);
    }
    hot_in.read((char*)&list_length, sizeof(list_length));
    for (std::size_t i=0; i<list_length; i++)
    {
      FACE_HORIZ_BOTTOM_NO_SLIP face(hot_in);
      face_horiz_list.push_back(face);
    }
    hot_in.read((char*)&list_length, sizeof(list_length));
    for (std::size_t i=0; i<list_length; i++)
    {
      FACE_HORIZ_BOTTOM_NO_DRAG face(hot_in);
      face_horiz_list.push_back(face);
    }
    hot_in.read((char*)&list_length, sizeof(list_length));
    for (std::size_t i=0; i<list_length; i++)
    {
      FACE_HORIZ_BOTTOM_QUAD_DRAG face(hot_in);
      face_horiz_list.push_back(face);
    }
    hot_in.read((char*)&list_length, sizeof(list_length));
    for (std::size_t i=0; i<list_length; i++)
    {
      FACE_HORIZ_BOTTOM_LIN_DRAG face(hot_in);
      face_horiz_list.push_back(face);
    }
#ifdef TEST_KEY
    hot_in.read((char*)&list_length, sizeof(list_length));
    for (std::size_t i=0; i<list_length; i++)
    {
      FACE_HORIZ_BOTTOM_TEST face(hot_in);
      face_horiz_list.push_back(face);
    }
#endif

    // Input vertical faces
    hot_in.read((char*)&list_length, sizeof(list_length));
    for (std::size_t i=0; i<list_length; i++)
    {
      FACE_VERT_INTERIOR face(hot_in);
      face_vert_list.push_back(face);
    }
    hot_in.read((char*)&list_length, sizeof(list_length));
    for (std::size_t i=0; i<list_length; i++)
    {
      FACE_VERT_LAND face(hot_in);
      face_vert_list.push_back(face);
    }
    hot_in.read((char*)&list_length, sizeof(list_length));
    for (std::size_t i=0; i<list_length; i++)
    {
      FACE_VERT_RIVER face(hot_in);
      face_vert_list.push_back(face);
    }
    hot_in.read((char*)&list_length, sizeof(list_length));
    for (std::size_t i=0; i<list_length; i++)
    {
      FACE_VERT_SEA face(hot_in);
      face_vert_list.push_back(face);
    }
    hot_in.read((char*)&list_length, sizeof(list_length));
    for (std::size_t i=0; i<list_length; i++)
    {
      FACE_VERT_RADIATION face(hot_in);
      face_vert_list.push_back(face);
    }
#ifdef TEST_KEY
    hot_in.read((char*)&list_length, sizeof(list_length));
    for (std::size_t i=0; i<list_length; i++)
    {
      FACE_VERT_UNKNOWN face(hot_in);
      face_vert_list.push_back(face);
    }
#endif
    // Output vertical split child index storage for vertical faces
    FACE_VERT::InputChildVert(hot_in);

    // Input 3D elements
    hot_in.read((char*)&list_length, sizeof(list_length));
    for (std::size_t i=0; i<list_length; i++)
    {
      PRISM_PARALLEL prism(hot_in);
      prism_list.push_back(prism);
    }
    hot_in.read((char*)&list_length, sizeof(list_length));
    for (std::size_t i=0; i<list_length; i++)
    {
      PRISM_NONPARALLEL prism(hot_in);
      prism_list.push_back(prism);
    }

    // Initialize storage for degrees of freedom
    Initialize_DOF();

    // Input primary varibles
    for (std::size_t i=0; i<d_var; i++)
      hot_in >> var[i];

    // Input recording stations info
    hot_in >> elev_stations;
    hot_in >> vel_stations;
    hot_in.close();
  }
  catch(std::exception& ex)
  {
    std::cerr << "File mesh_container.cpp: I/O error in configuration file" << ex.what() << '\n';
    std::exit(1);
  }

  // Set quadrature on 2D elements
  for (std::size_t i=0; i<triangle_2d_list.size(); i++)
  {
    triangle_2d_list[i].InitializeStorage();
    triangle_2d_list[i].Update_basis_quad();
  }

  // Set quadrature on 2D edges
  for (std::size_t i=0; i<edge_2d_list.size(); i++)
  {
    edge_2d_list[i].InitializeStorage();
    edge_2d_list[i].Update_basis_quad();
  }

  // Set quadrature on 3D elements
  for (std::size_t i=0; i<prism_list.size(); i++)
  {
    if (prism_list[i].IsActive())
      IncrementActiveElementCnt();
    prism_list[i].InitializeStorage();
    prism_list[i].Update_basis_quad();
  }

  // Set quadrature on horizontal faces
  for (std::size_t i=0; i<face_horiz_list.size(); i++)
  {
    face_horiz_list[i].InitializeStorage();
    face_horiz_list[i].Update_basis_quad();
  }

  // Set quadrature on vertical faces
  for (std::size_t i=0; i<face_vert_list.size(); i++)
  {
    face_vert_list[i].InitializeStorage();
    face_vert_list[i].Update_basis_quad();
  }
}
