#include <iostream>
#include <cassert>
#include <ctime>
#include <deque>
#include "general.H"
#include "basis.H"
#include "les_solver.H"
#include "initial_data.H"
#include "utility.H"
#include "runge_kutta.H"
#include "pde.H"
#include "mesh_container.H"

// Global variables
PROBLEM_DESCRIPTION problem_info;  // Storage for physical parameters
QUAD_CONTAINER quadrature_index;   // Quadrature objects

// 2D basis objects
BASIS_2D_CONST basis_2d_const;
BASIS_2D_LIN basis_2d_lin;
BASIS_2D_QUAD basis_2d_quad;
BASIS_2D *basis_2d[max_order+1]={&basis_2d_const, &basis_2d_lin, &basis_2d_quad};

// 3D basis objects
BASIS_REF_CONST basis_ref_const;
BASIS_REF_LIN basis_ref_lin;
BASIS_REF_QUAD basis_ref_quad;
BASIS_REF *basis_ref[max_order+1]={&basis_ref_const, &basis_ref_lin, &basis_ref_quad};
std::deque<BASIS_ZPHYS_CONST> basis_phys_const;
std::deque<BASIS_ZPHYS_LIN> basis_phys_lin;
std::deque<BASIS_ZPHYS_QUAD> basis_phys_quad;

// Global variables
DOF var_stage[max_stages][d_var];  // Global storage for degrees of freedom for Runge-Kutta stages
DOF var_rhs[d_var];                // Global storage for degrees of freedom for the right hand side
DOF var_flux[d_var][d_space];      // Global storage for degrees of freedom for fluxes
DOF var[d_var];                    // Global storage for degrees of freedom for unknowns

MESH mesh;                         // Mesh object

int main(int argc, char *argv[])
{
  const char config_default[]="utbest_config.inp";  // Name of default configuration file
  long_real t;
  std::time_t start_clock=std::time(0);

  // Read the configuration file and generate an initial 3D mesh
  switch(argc)
  {
  case 1: InputConfig(config_default, t);
    break;
  case 2: InputConfig(argv[1], t);
    break;
  default: std::cerr << "Wrong number of command line arguments!\n";
    std::exit(1);
  }

  long_real start_time=86400.*problem_info.start_time, dt=problem_info.dt;
  long_real end_time=start_time+86400.*problem_info.simul_length;

  // Create Runge-Kutta objects
  RUNGE_KUTTA rk_method(problem_info.runge_kutta_method);

  if (problem_info.hot_input==0)
  {
    Initialize_DOF();

    // Project initial surface elevation
    var[XI]=0.;
    IterateOverList(t, mesh.triangle_2d_list, initial_elevation, &TRIANGLE_2D_SW::Projection_elevation);

    // Project initial condition
    for (int i=0; i<d_var_3d_dt; i++)
      var[VAR_3D_DT_INDEX[i]]=0.;
    IterateOverList(start_time, mesh.prism_list, initial_condition, &PRISM_SW::Projection);

    // Multiply by an inverse of the mass matrix in case of a non-parallel prism
    for (std::size_t i=0; i<mesh.prism_nonparallel_list.size(); i++)
    {
      PRISM_NONPARALLEL& prism=mesh.prism_nonparallel_list[i];
      if (prism.IsActive())
        for (int j=0; j<d_var_3d_dt; j++)
          LU_substitution(prism.Var_dof(VAR_3D_DT_INDEX[j]).GetDim(),
          prism.mass_matrix(VAR_3D_DT_INDEX[j]), &prism.Var_dof(VAR_3D_DT_INDEX[j]));
    }
  }

  // Output the initial state of the system
  mesh.Output(start_time, 0);

  // Time stepping
  for (int end_ind=0; end_ind==0 && t<end_time; t+=dt)
  {
    // Update the time step
    if (dt>=end_time-t)
      dt=end_time-t, end_ind=1;
    // Update the ramp parameter if necessary
    if ((problem_info.ramp==true) && (t<start_time+86400.*problem_info.ramp_length))
      problem_info.ramp_value=(t-start_time)/(86400.*problem_info.ramp_length);
    else
      problem_info.ramp_value=1.;

    rk_method.time_stepping(t, dt);
    mesh.Output(t+dt, 3);  // Output change in the state variables
  }

  // Output the final state of the system
  mesh.Output(end_time, 1);
  std::cout << "Total CPU time = " << std::difftime(std::time(0),start_clock) <<" seconds" << std::endl;
  return 0;
}

